//============================================================================
// Copyright  Micromega Corporation 2005-2009. All rights reserved.
//
// @file:	fpu_i2c.c
// @target:	CCS PIC C Compiler
//
// This file provides support routines for using the uM-FPU V3.1 floating point
// coprocessor with the CCS PIC C compiler for PICmicro. It implements an I2C
// interface using the CCS I2C built-in functions. This file is added to the main
// program using an include statement.  The user must set the appropriate
// definitions for the interface pins (see examples below).
//
// @author Cam Thompson, Micromega Corporation
// @version
//	February 6, 2009
//	- modified for I2C
//	May 6, 2008
//	- updated for CCS PIC C compiler
//  February 17, 2005
//	- original version
//============================================================================

#include <fpu.h>

//-------------------- uM-FPU pin definitions --------------------------------
// Note: These pin definitions must be changed if other pins are used        //
//       for the I2C interface.                                              //
//-----------------------------------------------------------------------------
#use i2c(Master,fast,sda=PIN_C5,scl=PIN_C4)

#define FPU_WRITE_ADDR	0xC8			// FPU I2C device address (write)
#define	FPU_READ_ADDR	0xC9			// FPU I2C device address (read)

#define READ_DELAY		15				// minimum read setup delay

//-------------------- fpu_reset ---------------------------------------------

char fpu_reset()
{
	i2c_stop();
	i2c_start();
	i2c_write(FPU_WRITE_ADDR);
	i2c_write(1);
	i2c_write(0);
	i2c_stop();

	delay_ms(10);

	fpu_startWrite();
	i2c_write(SYNC);
	i2c_stop();

	return fpu_read();
}

//-------------------- fpu_startWrite, fpu_startRead -------------------------

void fpu_startWrite()
{
	i2c_start();
	i2c_write(FPU_WRITE_ADDR);
	i2c_write(0);
}

void fpu_startRead()
{
	i2c_start();
	i2c_write(FPU_READ_ADDR);
}

//-------------------- fpu_wait ----------------------------------------------

void fpu_wait(void)
{
	char bval;
	do
	{
		bval = fpu_read();
	}
	while (bval);
}

//-------------------- fpu_write ---------------------------------------------

void fpu_write(char bval)
{
	fpu_startWrite();
	i2c_write(bval);				// write one byte
	i2c_stop();
}

void fpu_write2(char bval1, char bval2)
{
	fpu_startWrite();
	i2c_write(bval1);				// write two bytes
	i2c_write(bval2);
	i2c_stop();
}

void fpu_write3(char bval1, char bval2, char bval3)
{
	fpu_startWrite();
	i2c_write(bval1);				// write three bytes
	i2c_write(bval2);
	i2c_write(bval3);
	i2c_stop();
}

void fpu_write4(char bval1, char bval2, char bval3, char bval4)
{
	fpu_startWrite();
	i2c_write(bval1);				// write four bytes
	i2c_write(bval2);
	i2c_write(bval3);
	i2c_write(bval4);
	i2c_stop();
}

void fpu_write5(char bval1, char bval2, char bval3, char bval4, char bval5)
{
	fpu_startWrite();
	i2c_write(bval1);				// write five bytes
	i2c_write(bval2);
	i2c_write(bval3);
	i2c_write(bval4);
	i2c_write(bval5);
	i2c_stop();
}

void fpu_write6(char bval1, char bval2, char bval3, char bval4, char bval5, char bval6)
{
	fpu_startWrite();
	i2c_write(bval1);				// write six bytes
	i2c_write(bval2);
	i2c_write(bval3);
	i2c_write(bval4);
	i2c_write(bval5);
	i2c_write(bval6);
	i2c_stop();
}

//-------------------- fpu_writeWord -----------------------------------------

void fpu_writeWord(long wval)
{
	union {
		long	wval;
		int		bval[2];
	} u;

	u.wval = wval;					// write 16-bit integer value
	fpu_startWrite();
	i2c_write(u.bval[1]);
	i2c_write(u.bval[0]);
	i2c_stop();
}

//-------------------- fpu_writeLong -----------------------------------------

void fpu_writeLong(int32 lval)
{
	union {
		int32	lval;
		int		bval[4];
	} u;

	u.lval = lval;					// write 32-bit integer value
	fpu_startWrite();
	i2c_write(u.bval[3]);
	i2c_write(u.bval[2]);
	i2c_write(u.bval[1]);
	i2c_write(u.bval[0]);
	i2c_stop();
}

//-------------------- fpu_writeFloat ----------------------------------------

void fpu_writeFloat(float fval)
{
	union {
		float	fval;
		int		bval[4];
	} u;

	u.fval = fval;					// write 32-bit floating point value
	fpu_startWrite();
	i2c_write(u.bval[0]);
	i2c_write(u.bval[1]);
	i2c_write(u.bval[2]);
	i2c_write(u.bval[3]);
	i2c_stop();
}

//-------------------- fpu_writeChar -----------------------------------------

void fpu_writeChar(char c)
{
	fpu_write(c);					// write character
}

//-------------------- fpu_writeString ---------------------------------------


void fpu_writeString(char *s)
{
	fpu_startWrite();				// write zero terminated string
	do
	{
		i2c_write(*s);
	} while (*s++);
	i2c_stop();
}

//-------------------- fpu_read ----------------------------------------------

char fpu_read()
{
	char bval;

	fpu_startWrite();
	fpu_startRead();				// read 8-bit byte
	bval = i2c_read(0);
	i2c_stop();
	return bval;
}

//-------------------- fpu_readWord ------------------------------------------

long fpu_readWord(void)
{
	union {
		int16	wval;
		int		bval[2];
	} u;

	fpu_startWrite();
	fpu_startRead();				// read 16-bit integer
	u.bval[1] = i2c_read();
	u.bval[0] = i2c_read(0);
	i2c_stop();
	return u.wval;
}

//-------------------- fpu_readLong ------------------------------------------

int32 fpu_readLong(void)
{
	union {
		int32	lval;
		int		bval[4];
	} u;

	fpu_startWrite();
	fpu_startRead();				// read 32-bit integer value
	u.bval[3] = i2c_read();
	u.bval[2] = i2c_read();
	u.bval[1] = i2c_read();
	u.bval[0] = i2c_read(0);
	i2c_stop();
	return u.lval;
}

//-------------------- fpu_readFloat -----------------------------------------

float fpu_readFloat(void)
{
	union {
		float	fval;
		int		bval[4];
	} u;

	fpu_startWrite();
	fpu_startRead();				// read 32-bit floating point value
	u.bval[0] = i2c_read();
	u.bval[1] = i2c_read();
	u.bval[2] = i2c_read();
	u.bval[3] = i2c_read(0);
	i2c_stop();
	return u.fval;
}

//-------------------- fpu_readString ----------------------------------------

void fpu_readString(char *s)
{

	fpu_startWrite();
	fpu_startRead();				// read zero-terminated string
	do {
		*s = i2c_read();
	} while (*s++);
	i2c_read(0);
	i2c_stop();
}

//-------------------- fpu_readStatus ----------------------------------------

int fpu_readStatus()
{
	int bval;

	fpu_startWrite();				// send read status instruction
	fpu_write(READSTATUS);
	fpu_startRead();
	bval = i2c_read(0);
	i2c_stop();
	return bval;
}

//-------------------- fpu_readDelay -----------------------------------------

void fpu_readDelay(void)
{
/*	Note: delay not required, since I2C overhead has sufficient delay
	delay_us(READ_DELAY);			// read setup delay
*/
}

//-------------------- print_version -----------------------------------------

void print_version()
{
	fpu_write(VERSION);				// print version string
	print_fpuString(READSTR);
}

//-------------------- print_float -------------------------------------------

void print_float(char format)
{
	fpu_write2(FTOA, format);		// print floating point string
	print_fpuString(READSTR);
}

//-------------------- print_long --------------------------------------------

void print_long(char format)
{
	fpu_write2(LTOA, format);		// print long integer string
	print_fpuString(READSTR);
}

//-------------------- print_fpuString ---------------------------------------

void print_fpuString(char opcode)
{
	char bval;

	fpu_wait();						// print FPU string
	fpu_startWrite();
	if (opcode != READSEL) opcode = READSTR;
	i2c_write(opcode);
	fpu_startRead();

	do {
		bval = i2c_read();
		if (bval) putc(bval);
	}
	while (bval);

	i2c_read(0);
	i2c_stop();
}

//-------------------- print_CRLF --------------------------------------------

void print_CRLF(void)
{
	putc('\r');						// print CR, LF
	putc('\n');
}
