//=============================================================================
// Copyright  Micromega Corporation 2005-2009. All rights reserved.
//
// @file:	fpu_spi.c
// @target:	CCS PIC C Compiler
//
// This file provides support routines for using the uM-FPU V1.0 floating point
// coprocessor with the CCS PIC C compiler for PICmicro. It implements the SPI
// interface in hardware and software (bit-bang). This file is added to the main
// program using an include statement.  The user must set the appropriate
// definitions for the interface pins (see examples below).
//
// @author Cam Thompson, Micromega Corporation
// @version
//	February 6, 2009
//	- updated comments
//	May 6, 2008
//	- updated for CCS PIC C compiler
//	February 17, 2005
//	- original version
//=============================================================================

#include <fpu.h>

//-------------------- uM-FPU pin definitions ---------------------------------

#define	FPU_SCK		PIN_C3				// SPI clock (FPU pin: SCLK)
#define	FPU_MISO	PIN_C4				// master in, slave out (FPU pin: SOUT)
#define	FPU_MOSI	PIN_C5				// master out, slave in (FPU pin: SIN)

//-------------------- uM-FPU definitions -------------------------------------

// if SPI_HARDWARE is not defined, the SPI signals are generated in software.
// if SPI_HARDWARE is defined, the hardware SPI uinit is used.
#define	SPI_HARDWARE	1				// use SPI hardware interface

// FOSC must be set to the clock frequency in hertz. (e.g. 20MHz = 20000000)
// FOSC is used to calculate the required delay to ensure that the minimum
// data period is1.6 usec.
#define FOSC 20000000					// clock speed in hertz

#ifdef SPI_HARDWARE
	#define DELAY_UNITS	(FOSC/10*4)		// delay units
	#if DELAY_UNITS < 8000000
		#define DATA_DELAY 1			// data delay
	#else
		#define DATA_DELAY (DELAY_UNITS/1000000 - 8) + 1
	#endif
#else
		#define DATA_DELAY 1			// data delay
#endif

#define READ_DELAY		15				// minimum read setup delay

//-------------------- fpu_reset ----------------------------------------------

char fpu_reset(void) {
	int i;

#ifdef SPI_HARDWARE
	setup_spi(SPI_SS_DISABLED);			// turn off SPI hardware
#endif

	output_low(FPU_SCK);				// send reset pulses
	output_high(FPU_MOSI);
	for (i = 0; i < 80; i++) {
		output_high(FPU_SCK);
		output_low(FPU_SCK);
	}
	output_low(FPU_MOSI);
	input(FPU_MISO);

	delay_ms(10);						// wait for reset delay

#ifdef SPI_HARDWARE
	setup_spi(SPI_MASTER | SPI_L_TO_H | SPI_CLK_DIV_4 | SPI_XMIT_L_TO_H);
#endif

	fpu_write2(PICMODE, SYNC);			// set to PIC mode, and send SYNC
	fpu_readDelay();					// to confirm communications
	return (fpu_read());
}

//-------------------- fpu_wait -----------------------------------------------

void fpu_wait(void) {
	while (input(FPU_MISO)) ;			// wait for FPU ready status
}

//-------------------- fpu_write ----------------------------------------------

void fpu_write(char bval) {

	#ifdef SPI_HARDWARE
		spi_write(bval);
	#else
		int i;
		for (i = 0; i < 8; i++) {			// repeat for 8 bits
			output_low(FPU_SCK);			// set clock low
			if (bval & 0x80)				// send next bit to FPU_MOSI
				output_high(FPU_MOSI);
			else
				output_low(FPU_MOSI);
			bval <<= 1;
			output_high(FPU_SCK);			// set clock high
		}
		output_low(FPU_SCK);				// set clock low at end of byte
	#endif
	delay_cycles(DATA_DELAY);			// minimum data period = 1.6 usec
}

void fpu_write2(char bval1, char bval2) {
	fpu_write(bval1);					// write two bytes
	fpu_write(bval2);
}

void fpu_write3(char bval1, char bval2, char bval3) {
	fpu_write(bval1);					// write three bytes
	fpu_write(bval2);
	fpu_write(bval3);
}

void fpu_write4(char bval1, char bval2, char bval3, char bval4) {
	fpu_write(bval1);					// write four bytes
	fpu_write(bval2);
	fpu_write(bval3);
	fpu_write(bval4);
}

void fpu_write5(char bval1, char bval2, char bval3, char bval4, char bval5) {
	fpu_write(bval1);					// write five bytes
	fpu_write(bval2);
	fpu_write(bval3);
	fpu_write(bval4);
	fpu_write(bval5);
}

void fpu_write6(char bval1, char bval2, char bval3, char bval4, char bval5, char bval6) {
	fpu_write(bval1);					// write six bytes
	fpu_write(bval2);
	fpu_write(bval3);
	fpu_write(bval4);
	fpu_write(bval5);
	fpu_write(bval6);
}

//-------------------- fpu_writeWord ------------------------------------------

void fpu_writeWord(long wval) {
	fpu_write(wval>>8);					// write 16-bit integer value
	fpu_write(wval & 0xFF);
}

//-------------------- fpu_writeLong ----------------------------------------

void fpu_writeLong(int32 lval) {
	union {
		int32	lval;
		int		bval[4];
	} u;
	u.lval = lval;
	fpu_write(u.bval[3]);				// write 32-bit integer value
	fpu_write(u.bval[2]);
	fpu_write(u.bval[1]);
	fpu_write(u.bval[0]);
}

//-------------------- fpu_writeFloat -----------------------------------------

void fpu_writeFloat(float fval) {
	union {
		float	fval;
		int		bval[4];
	} u;
	u.fval = fval;
	fpu_write(u.bval[0]);				// write 32-bit floating point value
	fpu_write(u.bval[1]);
	fpu_write(u.bval[2]);
	fpu_write(u.bval[3]);
}

//-------------------- fpu_writeChar -----------------------------------------

void fpu_writeChar(char c) {
	fpu_write(c);						// write character
}

//-------------------- fpu_writeString ---------------------------------------


void fpu_writeString(char *s) {
	do {
		fpu_write(*s);					// write zero terminated string
	} while (*s++);
}

//-------------------- fpu_read -----------------------------------------------

char fpu_read() {
	#ifdef SPI_HARDWARE
		int n;
		n = spi_read(0);					// read byte using SPI hardware

	#else
		int n, i;
		n = 0;								// read byte using software (bit-bang)
		for (i = 0; i < 8; i++) {
	 		output_low(FPU_SCK);			// set clock low
			n <<= 1;						// read next bit from FPU_MISO
			if (input(FPU_MISO)) n |= 1;
			output_high(FPU_SCK);			// set clock high
		}
	#endif
	delay_cycles(DATA_DELAY);			// minimum data period = 1.6 usec
	return n;
}

//-------------------- fpu_readWord -------------------------------------------

long fpu_readWord(void) {				// read 16-bit integer value
	return (fpu_read() << 8) + fpu_read();
}

//-------------------- fpu_readLong -------------------------------------------

int32 fpu_readLong(void) {
	union {
		int32	lval;
		int		bval[4];
	} u;
	u.bval[3] = fpu_read();				// read 32-bit integer value
	u.bval[2] = fpu_read();
	u.bval[1] = fpu_read();
	u.bval[0] = fpu_read();
	return u.lval;
}

//-------------------- fpu_readFloat ------------------------------------------

float fpu_readFloat(void) {
	union {
		float	fval;
		int		bval[4];
	} u;
	u.bval[0] = fpu_read();				// read 32-bit floating point value
	u.bval[1] = fpu_read();
	u.bval[2] = fpu_read();
	u.bval[3] = fpu_read();
	return u.fval;
}

//-------------------- fpu_readString ---------------------------------------

void fpu_readString(char *s) {
	do {
		*s = fpu_read();				// read zero-terminated string
	} while (*s++);
}

//-------------------- fpu_readStatus -----------------------------------------

int fpu_readStatus() {
	fpu_write(READSTATUS);				// send read status instruction
	fpu_readDelay();
	return fpu_read();					// return status byte
}

//-------------------- fpu_readDelay ------------------------------------------

void fpu_readDelay(void) {
	delay_us(READ_DELAY);				// read setup delay
}

//-------------------- print_version ------------------------------------------

void print_version() {
	fpu_write(VERSION);					// print version string
	print_fpuString(READSTR);
}

//-------------------- print_float --------------------------------------------

void print_float(char format) {
	fpu_write2(FTOA, format);			// print floating point string
	print_fpuString(READSTR);
}

//-------------------- print_long ---------------------------------------------

void print_long(char format) {
	fpu_write2(LTOA, format);			// print long integer string
	print_fpuString(READSTR);
}

//-------------------- print_fpuString ----------------------------------------

void print_fpuString(char opcode) {
	char bval;

	if (opcode != READSEL) opcode = READSTR;
	fpu_wait();							// wait until FPU is ready
	fpu_write(opcode);					// send read string opcode
	fpu_readDelay();

	do {							// loop until zero terminator
		bval = fpu_read();				// get the next byte
		if (bval) putc(bval);			// print if non-zero
	} while (bval);
}

//-------------------- print_CRLF ---------------------------------------------

void print_CRLF(void) {
	putc('\r');							// print CR, LF
	putc('\n');
}
